//----------------------------------------------------------------------------
//~Author: Fabio Arciniegas A.
//~Name: JPG2QuadTree.xml
//~Description: Converts a Black and White JPG image into an XML representation
//              suitable for the program com.faa.sax.tutorial.QuadTreeViewer
//
//Version $Id$

//----------------------------------------------------------------------------
//             This code has been formated to fit in 80 columns
//----------------------------------------------------------------------------
package com.faa.graphics.util;


import java.io.                  File;
import java.io.                  FileOutputStream;
import java.io.                  FileInputStream;
import java.io.                  PrintStream;
import java.io.                  OutputStream;
import com.sun.image.codec.jpeg. JPEGCodec;
import com.sun.image.codec.jpeg. JPEGImageDecoder;
import java.awt.                 Image;
import java.awt.image.           PixelGrabber;
import java.awt.image.           ImageObserver;



public class JPG2QuadTree
{

    // ******************************************                           Main

    /**
     * Receive the name of the jpg to convert, plus the optional output filename and 
     * convert the JPG into a QuadTree
     * @param args only two arguments are allowed: the name of the jpg and the output
     *             file name. 
     */
    public static void main(String args[])
    {
	if(args.length > 2 || args.length < 1)
	{
	    usage();
	    return;
	}
	

	if(args.length == 2) // if there is an output file, use it
	{
	    File outputFile = new File(args[1]);
	    
	    try
	    {
		FileOutputStream fo = new FileOutputStream(outputFile);
		out = new PrintStream((OutputStream)fo);
            }
	    catch(java.io.IOException e)
	    {
		System.out.println("Can't write to " + args[1] + ".Terminating");
		System.out.println(e);
	    }
	}
	else // else, use standard output
	    out = System.out;
	// Now, get the image from the file
	processImage(args[0]);


    }

   /** 
      Get the source image from the file 
      <em>precondition:</em> the file exists, can be read, and is a B/W JPEG
      @param inputFile the name of the file
   */
    public static void processImage(String inputFile)
    {
	File infile = new File(inputFile);
	JPEGCodec jpegCodec = null;
	JPEGImageDecoder jpegSrc = null;

	try
	{
	    FileInputStream in = new FileInputStream(infile);
	    jpegSrc = jpegCodec.createJPEGDecoder(in);
	    source = jpegSrc.decodeAsBufferedImage();
	}
	catch(java.io.IOException e)
	{
	    System.out.println("Unable to read " + inputFile + ":" + e);
	    return;
	}
	catch(com.sun.image.codec.jpeg.ImageFormatException e)
	{
	    System.out.println("Malformed JPG (" + inputFile + "):" + e);	    
	    return;
	}
	// we now have an image. we are ready to process.
	int h = source.getHeight(null);
	int w = source.getWidth(null);
	int size = h < w ? h : w; // clip it to a square
	// Make sure we deal with a size which is a power of 2
	// after all, this is a QuadTree!
	imageSize = 2;
	while(imageSize < size)
	    imageSize = imageSize * 2;

	size = imageSize;
	pixels = new int[size * size];
        pg = new PixelGrabber(source, 0, 0, size, size, pixels, 0, size);
	// the pixels variable is now totally populated by the image info

	out.println("<?xml version='1.0'?>");
	out.println("<quadtree size='" + size + "'>");
	out.println("<!-- This file was created with com.faa.graphics.JPG2QuadTree -->");
	createQuadTreeNode(0,0,size); 
	out.println("</quadtree>");
    }
    
 
  /** 
      Any pixel on the image which is not RGB == (255,255,255) is treated as
      a black pixel. This method tells whether a pixel is white or not
      @param pixel the pixel to treat
      @return whether the pixel is white or not
   */
    public static boolean isWhite(int pixel) {
	int red   = (pixel >> 16) & 0xff;
	int green = (pixel >>  8) & 0xff;
	int blue  = pixel & 0xff;
	return ((red | green | blue) != 0);
    }

  /** 
      Create a node on the xml file, representing the given portion of the image.
      Note that this may imply calling this function recursively several times
      @param x the x coordinate for the square portion
      @param y the y coordinate for the square portion
      @param size the side (witdth/height) of the square portion
   */
    public static void createQuadTreeNode(int x, int y, int size) {
	int h,w,offset; 
	h = w = size;
	if(size < 1) 
	    return;
	//offset will position us at the beginning of the square in the
	//pixels array
	offset = y*imageSize + x;
	try {
	    pg.grabPixels();
	} catch (InterruptedException e) {
	    System.err.println("interrupted waiting for pixels!");
	    return;
	}
	if ((pg.getStatus() & ImageObserver.ABORT) != 0) {
	    System.err.println("image fetch aborted or errored");
	    return;
	}
	boolean allWhite = true;
	boolean allBlack = true;	


	for (int j = 0; j < h && allWhite; j++) {
	    for (int i = 0; i < w && allWhite; i++) {
		// Test for padding area 
		allWhite = isWhite(pixels[offset + (j * imageSize + i)]);
	    }
	}
	if(allWhite)
	{
	    out.println("<node value='WHITE'/>");
	    return;
	}

	for (int j = 0; j < h && allBlack; j++) {
	    for (int i = 0; i < w && allBlack; i++) {
		allBlack = !isWhite(pixels[offset +(j * imageSize + i)]);
	    }
	}
	if(allBlack)
	{
	    out.println("<node value='BLACK'/>");
	    return;
	}
	// the image portion contains mixed black and white pixels
        out.println("<node value='GRAY'>");
	createQuadTreeNode(x+size/2,y,size/2);
	createQuadTreeNode(x,y,size/2);
	createQuadTreeNode(x,y+size/2,size/2);
	createQuadTreeNode(x+size/2,y+size/2,size/2);
        out.println("</node>");
    }


   /** 
      Simply output the usage rules
   */
    public static void usage()
    {
	System.out.println("Usage: java com.faa.graphics.util.JPG2QuadTree " + 
                           "jpegFile [outputFile]");
    }

// ******************************************                          Atributes

    static PrintStream out;
    static Image source;
    static int[] pixels;
    static int imageSize;
    static PixelGrabber pg;
}
 






